/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file video-player.cpp
@ingroup ExamplePrograms
@brief Simple example of image input/output from a video file, to a canvas window

This program will receive a path or url to a video file, and then will read each frame and show each YUV channel in a canvas window.

@section Usage
Once compiled, the program should be used with this command line:
\code
	./video-player --URL=<path or url to video>
\endcode

@param URL Video source identificator. See dynamic property URL for @ref QVMPlayerCamera.
@author PARP Research Group. University of Murcia, Spain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream>

#include <qvcore/qvapplication.h>
#include <qvcameras/qvmplayercamera.h>
#include <qvgui/qvgui.h>

/// \cond
class PlayerWorker: public QVWorker
	{
	public:
		PlayerWorker(QString name): QVWorker(name)
			{
			addProperty< QVImage<uChar,1> >("Input image Y", inputFlag);
			addProperty< QVImage<uChar,1> >("Input image U", inputFlag);
			addProperty< QVImage<uChar,1> >("Input image V", inputFlag);

			addProperty< QVImage<uChar,1> >("Output image Y", outputFlag);
			addProperty< QVImage<uChar,1> >("Output image U", outputFlag);
			addProperty< QVImage<uChar,1> >("Output image V", outputFlag);
			}

		void iterate()
			{
			QVImage<uChar, 1>	imageY = getPropertyValue< QVImage<uChar,1> >("Input image Y"),
						imageU = getPropertyValue< QVImage<uChar,1> >("Input image U"),
						imageV = getPropertyValue< QVImage<uChar,1> >("Input image V");

			setPropertyValue< QVImage<uChar,1> >("Output image Y", imageY);
			setPropertyValue< QVImage<uChar,1> >("Output image U", imageU);
			setPropertyValue< QVImage<uChar,1> >("Output image V", imageV);
			}
	};

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		"Example program for QVision library. Shows the content of a video file."
		);

	QVMPlayerCamera camera("Video");
	PlayerWorker worker("Video player worker");
	camera.link(&worker,"Input image Y", "Input image U", "Input image V");

	QVGUI interface;

	QVImageCanvas imageCanvasY("Output image Y");
	imageCanvasY.linkProperty(worker, "Output image Y");

	QVImageCanvas imageCanvasU("Output image U");
	imageCanvasU.linkProperty(worker, "Output image U");

	QVImageCanvas imageCanvasV("Output image V");
	imageCanvasV.linkProperty(worker, "Output image V");

	return app.exec();
	}

/// \endcond
