/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvimage.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <QDebug>
#include <qvcore/qvimage.h>

template <> const char * QVImage<uChar,1>::getTypeQString()	const	{ return "QVImage<uChar,1>"; }
template <> const char * QVImage<uChar,3>::getTypeQString()	const	{ return "QVImage<uChar,3>"; }
template <> const char * QVImage<sShort,1>::getTypeQString()	const	{ return "QVImage<sShort,1>"; }
template <> const char * QVImage<sShort,3>::getTypeQString()	const	{ return "QVImage<sShort,3>"; }
template <> const char * QVImage<sFloat,1>::getTypeQString()	const	{ return "QVImage<sFloat,1>"; }
template <> const char * QVImage<sFloat,3>::getTypeQString()	const	{ return "QVImage<sFloat,3>"; }

// Copy constructors
#define	CREATE_COPY_CONSTRUCTOR(TYPE, C)							\
template <> QVImage<TYPE, C>::QVImage(QVImage<TYPE, C> const &img):QVGenericImage(img)		\
	{											\
	imageBuffer = img.imageBuffer;								\
	setROI(img.getROI()); setAnchor(img.getAnchor());					\
	}

CREATE_COPY_CONSTRUCTOR(uChar, 1);
CREATE_COPY_CONSTRUCTOR(uChar, 3);
CREATE_COPY_CONSTRUCTOR(sShort, 1);
CREATE_COPY_CONSTRUCTOR(sShort, 3);
CREATE_COPY_CONSTRUCTOR(sFloat, 1);
CREATE_COPY_CONSTRUCTOR(sFloat, 3);

// Conversion operator
#define	CREATE_COPY_OPERATOR(TYPE, C)							\
template <> QVImage<TYPE,C> & QVImage<TYPE, C>::operator=(const QVImage<TYPE, C> &img) 	\
	{										\
	imageBuffer = img.imageBuffer;							\
	setROI(img.getROI()); setAnchor(img.getAnchor());				\
	return *this;									\
	}

CREATE_COPY_OPERATOR(uChar, 1);
CREATE_COPY_OPERATOR(uChar, 3);
CREATE_COPY_OPERATOR(sShort, 1);
CREATE_COPY_OPERATOR(sShort, 3);
CREATE_COPY_OPERATOR(sFloat, 1);
CREATE_COPY_OPERATOR(sFloat, 3);

//////////////////////////////////////////////////////////////////
// QVIPP FUNCTIONS
#include <qvipp/qvipp.h>

// Conversion constructors
#define	CREATE_CONVERT_CONSTRUCTOR(TYPE1, C1, TYPE2, C2)					\
template <> QVImage<TYPE1, C1>::QVImage(QVImage<TYPE2, C2> const &img):QVGenericImage(img)	\
	{											\
	imageBuffer = new QVImageBuffer<TYPE1, C1>(img.getCols(), img.getRows());		\
	setAnchor(img.getROI().x(),img.getROI().y());						\
	Convert(img, *this);								\
 	setAnchor(img.getAnchor());								\
	}

CREATE_CONVERT_CONSTRUCTOR(uChar, 1, sShort, 1);
CREATE_CONVERT_CONSTRUCTOR(uChar, 1, sFloat, 1);
CREATE_CONVERT_CONSTRUCTOR(sShort, 1, uChar, 1);
CREATE_CONVERT_CONSTRUCTOR(sShort, 1, sFloat, 1);
CREATE_CONVERT_CONSTRUCTOR(sFloat, 1, uChar, 1);
CREATE_CONVERT_CONSTRUCTOR(sFloat, 1, sShort, 1);

CREATE_CONVERT_CONSTRUCTOR(uChar, 3, sShort, 3);
CREATE_CONVERT_CONSTRUCTOR(uChar, 3, sFloat, 3);
CREATE_CONVERT_CONSTRUCTOR(sShort, 3, uChar, 3);
CREATE_CONVERT_CONSTRUCTOR(sShort, 3, sFloat, 3);
CREATE_CONVERT_CONSTRUCTOR(sFloat, 3, uChar, 3);
CREATE_CONVERT_CONSTRUCTOR(sFloat, 3, sShort, 3);

#define	CREATE_CONVERT_CONSTRUCTOR_C3_C1(TYPE)							\
template <> QVImage<TYPE, 1>::QVImage(QVImage<TYPE, 3> const &img):QVGenericImage(img)		\
	{											\
	imageBuffer = new QVImageBuffer<TYPE, 1>(img.getCols(), img.getRows());			\
	setAnchor(img.getROI().x(),img.getROI().y());						\
	RGBToGray(img, *this);								\
 	setAnchor(img.getAnchor());								\
	}

CREATE_CONVERT_CONSTRUCTOR_C3_C1(uChar);
CREATE_CONVERT_CONSTRUCTOR_C3_C1(sShort);
CREATE_CONVERT_CONSTRUCTOR_C3_C1(sFloat);

#define	CREATE_CONVERT_CONSTRUCTOR_C1_C3(TYPE)							\
template <> QVImage<TYPE, 3>::QVImage(QVImage<TYPE, 1> const &img):QVGenericImage(img)		\
	{											\
	imageBuffer = new QVImageBuffer<TYPE, 3>(img.getCols(), img.getRows());			\
	setAnchor(img.getROI().x(),img.getROI().y());						\
	Copy(img, *this);								\
 	setAnchor(img.getAnchor());								\
	}

CREATE_CONVERT_CONSTRUCTOR_C1_C3(uChar);
CREATE_CONVERT_CONSTRUCTOR_C1_C3(sShort);
CREATE_CONVERT_CONSTRUCTOR_C1_C3(sFloat);


#define CREATE_OPERATOR(NAME, OPERATOR, TYPE, C)						\
template <> QVImage<TYPE, C> QVImage<TYPE,C>::OPERATOR(const QVImage<TYPE, C> &img) const	\
	{											\
	QVImage<TYPE, C> result = *this;							\
	NAME(*this, img, result);							\
	return result;										\
	}

CREATE_OPERATOR(Add, operator+, uChar, 1);
CREATE_OPERATOR(Mul, operator*, uChar, 1);
CREATE_OPERATOR(Sub, operator-, uChar, 1);
CREATE_OPERATOR(Div, operator/, uChar, 1);

CREATE_OPERATOR(Add, operator+, uChar, 3);
CREATE_OPERATOR(Mul, operator*, uChar, 3);
CREATE_OPERATOR(Sub, operator-, uChar, 3);
CREATE_OPERATOR(Div, operator/, uChar, 3);

CREATE_OPERATOR(Add, operator+, sShort, 1);
CREATE_OPERATOR(Mul, operator*, sShort, 1);
CREATE_OPERATOR(Sub, operator-, sShort, 1);
CREATE_OPERATOR(Div, operator/, sShort, 1);

CREATE_OPERATOR(Add, operator+, sShort, 3);
CREATE_OPERATOR(Mul, operator*, sShort, 3);
CREATE_OPERATOR(Sub, operator-, sShort, 3);
CREATE_OPERATOR(Div, operator/, sShort, 3);


#define CREATE_CONST_OPERATOR(NAME, OPERATOR, TYPE, C)				\
template <> QVImage<TYPE, C> QVImage<TYPE,C>::OPERATOR(const TYPE value) const	\
	{									\
	QVImage<TYPE, C> result = *this;					\
	NAME(*this, result, value);					\
	return result;								\
	}

CREATE_CONST_OPERATOR(AddC, operator+, uChar, 1);
CREATE_CONST_OPERATOR(AddC, operator+, sShort, 1);
CREATE_CONST_OPERATOR(AddC, operator+, sFloat, 1);

CREATE_CONST_OPERATOR(MulC, operator*, uChar, 1);
CREATE_CONST_OPERATOR(MulC, operator*, sShort, 1);
CREATE_CONST_OPERATOR(MulC, operator*, sFloat, 1);

CREATE_CONST_OPERATOR(SubC, operator-, uChar, 1);
CREATE_CONST_OPERATOR(SubC, operator-, sShort, 1);
CREATE_CONST_OPERATOR(SubC, operator-, sFloat, 1);

CREATE_CONST_OPERATOR(DivC, operator/, uChar, 1);
CREATE_CONST_OPERATOR(DivC, operator/, sShort, 1);
CREATE_CONST_OPERATOR(DivC, operator/, sFloat, 1);

CREATE_CONST_OPERATOR(LShiftC, operator<<, uChar, 1);
CREATE_CONST_OPERATOR(RShiftC, operator>>, uChar, 1);

CREATE_CONST_OPERATOR(AndC, operator&, uChar, 1);
CREATE_CONST_OPERATOR(OrC, operator|, uChar, 1);
CREATE_CONST_OPERATOR(XorC, operator^, uChar, 1);

#define CREATE_NOT_OPERATOR(NAME, OPERATOR, TYPE, C)		\
template <> QVImage<TYPE, C> QVImage<TYPE,C>::OPERATOR() const	\
	{							\
	QVImage<TYPE, C> result = *this;			\
	NAME(*this, result);				\
	return result;						\
	}

CREATE_NOT_OPERATOR(Not, operator!, uChar, 1);
CREATE_NOT_OPERATOR(Not, operator!, uChar, 3);

#define CREATE_COMPARE_OPERATOR(NAME, OPERATOR, CMP, TYPE, C)				\
template <> QVImage<uChar> QVImage<TYPE,C>::OPERATOR(const QVImage<TYPE, C> &img) const	\
	{										\
	QVImage<uChar> result(getCols(), getRows());					\
	NAME(*this, img, result, CMP);						\
	return result;									\
	}

// ippcmpLess
CREATE_COMPARE_OPERATOR(Compare, operator<, ippCmpLess, uChar, 1);
CREATE_COMPARE_OPERATOR(Compare, operator<, ippCmpLess, sShort, 1);
CREATE_COMPARE_OPERATOR(Compare, operator<, ippCmpLess, sFloat, 1);

CREATE_COMPARE_OPERATOR(Compare, operator<, ippCmpLess, uChar, 3);
CREATE_COMPARE_OPERATOR(Compare, operator<, ippCmpLess, sShort, 3);
CREATE_COMPARE_OPERATOR(Compare, operator<, ippCmpLess, sFloat, 3);

// ippCmpGreater
CREATE_COMPARE_OPERATOR(Compare, operator>, ippCmpGreater, uChar, 1);
CREATE_COMPARE_OPERATOR(Compare, operator>, ippCmpGreater, sShort, 1);
CREATE_COMPARE_OPERATOR(Compare, operator>, ippCmpGreater, sFloat, 1);

CREATE_COMPARE_OPERATOR(Compare, operator>, ippCmpGreater, uChar, 3);
CREATE_COMPARE_OPERATOR(Compare, operator>, ippCmpGreater, sShort, 3);
CREATE_COMPARE_OPERATOR(Compare, operator>, ippCmpGreater, sFloat, 3);

// ippcmpLessEq
CREATE_COMPARE_OPERATOR(Compare, operator<=, ippCmpLessEq, uChar, 1);
CREATE_COMPARE_OPERATOR(Compare, operator<=, ippCmpLessEq, sShort, 1);
CREATE_COMPARE_OPERATOR(Compare, operator<=, ippCmpLessEq, sFloat, 1);

CREATE_COMPARE_OPERATOR(Compare, operator<=, ippCmpLessEq, uChar, 3);
CREATE_COMPARE_OPERATOR(Compare, operator<=, ippCmpLessEq, sShort, 3);
CREATE_COMPARE_OPERATOR(Compare, operator<=, ippCmpLessEq, sFloat, 3);

// ippCmpGreaterEq
CREATE_COMPARE_OPERATOR(Compare, operator>=, ippCmpGreaterEq, uChar, 1);
CREATE_COMPARE_OPERATOR(Compare, operator>=, ippCmpGreaterEq, sShort, 1);
CREATE_COMPARE_OPERATOR(Compare, operator>=, ippCmpGreaterEq, sFloat, 1);

CREATE_COMPARE_OPERATOR(Compare, operator>=, ippCmpGreaterEq, uChar, 3);
CREATE_COMPARE_OPERATOR(Compare, operator>=, ippCmpGreaterEq, sShort, 3);
CREATE_COMPARE_OPERATOR(Compare, operator>=, ippCmpGreaterEq, sFloat, 3);


// Set FUNCTION
#define	CREATE_SET_FUNCTION_C1(TYPE)				\
template <> void QVImage<TYPE>::set(TYPE c1, TYPE, TYPE)	\
	{ Set(*this, c1); }

CREATE_SET_FUNCTION_C1(uChar);
CREATE_SET_FUNCTION_C1(sShort);
CREATE_SET_FUNCTION_C1(sFloat);

#define	CREATE_SET_FUNCTION_C3(TYPE)					\
template <> void QVImage<TYPE,3>::set(TYPE c1, TYPE c2, TYPE c3)	\
	{ Set(*this, c1, c2, c3); }

CREATE_SET_FUNCTION_C3(uChar);
CREATE_SET_FUNCTION_C3(sShort);
CREATE_SET_FUNCTION_C3(sFloat);


//////////////
// operator =
#define	CREATE_CONVERT_OPERATOR(TYPE1, TYPE2, C)						\
template <> QVImage<TYPE1, C> & QVImage<TYPE1, C>::operator=(const QVImage<TYPE2, C> &img)	\
	{											\
	imageBuffer = new QVImageBuffer<TYPE1, C>(img.getCols(), img.getRows());		\
	setAnchor(img.getROI().x(),img.getROI().y());						\
	Convert(img, *this);								\
 	setAnchor(img.getAnchor());								\
	return *this;										\
	}

CREATE_CONVERT_OPERATOR(uChar, sFloat, 1);
CREATE_CONVERT_OPERATOR(uChar, sShort, 1);
CREATE_CONVERT_OPERATOR(sShort, uChar, 1);
CREATE_CONVERT_OPERATOR(sShort, sFloat, 1);
CREATE_CONVERT_OPERATOR(sFloat, uChar, 1);
CREATE_CONVERT_OPERATOR(sFloat, sShort, 1);

CREATE_CONVERT_OPERATOR(uChar, sFloat, 3);
CREATE_CONVERT_OPERATOR(uChar, sShort, 3);
CREATE_CONVERT_OPERATOR(sShort, uChar, 3);
CREATE_CONVERT_OPERATOR(sShort, sFloat, 3);
CREATE_CONVERT_OPERATOR(sFloat, uChar, 3);
CREATE_CONVERT_OPERATOR(sFloat, sShort, 3);

#define	CREATE_CONVERT_OPERATOR_C3_C1(TYPE1, C1, TYPE2, C2)					\
template <> QVImage<TYPE1, C1> & QVImage<TYPE1, C1>::operator=(const QVImage<TYPE2, C2> &img)	\
	{											\
	imageBuffer = new QVImageBuffer<TYPE1, C1>(img.getCols(), img.getRows());		\
	setAnchor(img.getROI().x(),img.getROI().y());						\
	RGBToGray(img, *this);								\
 	setAnchor(img.getAnchor());								\
	return *this;										\
	}

CREATE_CONVERT_OPERATOR_C3_C1(uChar, 1, uChar, 3);

#define	CREATE_CONVERT_OPERATOR_C1_C3(TYPE1, C1, TYPE2, C2)					\
template <> QVImage<TYPE1, C1> & QVImage<TYPE1, C1>::operator=(const QVImage<TYPE2, C2> &img)	\
	{											\
	imageBuffer = new QVImageBuffer<TYPE1, C1>(img.getCols(), img.getRows());		\
	setAnchor(img.getROI().x(),img.getROI().y());						\
	Copy(img, *this);								\
 	setAnchor(img.getAnchor());								\
	return *this;										\
	}

CREATE_CONVERT_OPERATOR_C1_C3(uChar, 3, uChar, 1);

